# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2019-2020)
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""HTTP request method with end-to-end ECP authentication
"""

from __future__ import absolute_import

try:
    from urllib.request import Request
except ImportError:  # python < 3
    from urllib2 import Request

from .cookies import (
    COOKIE_FILE,
    ECPCookieJar,
    has_session_cookies,
    load_cookiejar,
)
from .ecp import authenticate
from .http import build_verified_opener
from .kerberos import has_credential
from .utils import format_endpoint_url

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def request(
        url,
        username=None,
        endpoint=None,
        kerberos=None,
        cookiejar=None,
        cookiefile=COOKIE_FILE,
        debug=False,
        store_session_cookies=False,
):
    """Request the given URL using ECP shibboleth authentication

        >>> from ligo.org import request
        >>> response = request(myurl)
        >>> print(response.read())

    Parameters
    ----------
    url : `str`
        URL path for request

    endpoint : `str`, optional
        ECP endpoint URL for request

    kerberos : `bool`, optional
        use existing kerberos credential for login, default is to try, but
        fall back to username/password prompt

    debug : `bool`, optional, default: `False`
        query in verbose debugging mode

    Returns
    -------
    response : `http.client.HTTPResponse`
        the response from the URL
    """

    # read existing cookies
    if cookiejar is None:
        cookiejar = load_cookiejar(cookiefile, strict=False)
    reuse = has_session_cookies(cookiejar, url)

    # -- authenticate ---------------------------

    if not reuse:
        if kerberos is None:  # guess availability of kerberos
            kerberos = has_credential()
        try:
            endpoint = format_endpoint_url(endpoint, kerberos=kerberos)
        except AttributeError:
            raise ValueError(
                "a valid endpoint name or URL is required when "
                "no existing cookie can be used"
            )
        authenticate(
            endpoint,
            username=username,
            spurl=url,
            kerberos=kerberos,
            cookiejar=cookiejar,
            debug=debug,
        )

        # cache cookies for next time (only if using our fancy jar)
        if isinstance(cookiejar, ECPCookieJar) and cookiefile is not None:
            cookiejar.save(
                cookiefile,
                ignore_discard=store_session_cookies,
                ignore_expires=store_session_cookies,
            )

    # -- actually send GET ----------------------

    myheaders = {'Accept': 'text/*'}    # allow any text mime not only html
    request = Request(url=url, headers=myheaders)
    opener = build_verified_opener(cookiejar=cookiejar, debug=debug)
    response = opener.open(request)
    return response



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2019-2020)
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Create an X.509 certificate using ECP authentication.
"""

from __future__ import print_function

import argparse
import os
import sys

from OpenSSL import crypto

from ..x509 import (
    check_cert,
    get_cert,
    get_x509_proxy_path,
    load_cert,
    print_cert_info,
    write_cert,
)
from .utils import (
    ArgumentParser,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

EPILOG = r"""
Examples:

    $ ecp-get-cert 'My Institution' user.name

to authenticate with a password prompt, or

    $ ecp-get-cert 'My Institution' --kerberos

to reuse an existing kerberos (``kinit``) credential.

The identitity provider name can be given in a number of ways, so long as the
argument uniquely identifies a provider.  For example, the following are all
equivalent:

    $ ecp-get-cert 'Cardiff University' user.name
    $ ecp-get-cert Cardiff user.name
    $ ecp-get-cert idp.cf.ac.uk user.name

Environment:

X509_USER_PROXY:
    the default path for the credential file
"""

MANPAGE = [
    {'heading': 'environment',
     'content': """
.TP
.B "X509_USER_PROXY"
The default path for the credential file """,
     },
]


def create_parser():
    """Create a command-line argument parser

    Returns
    -------
    parser : `argparse.ArgumentParser`
    """
    parser = ArgumentParser(
        description=__doc__,
        epilog=EPILOG,
        prog="ecp-get-cert",
        manpage=MANPAGE,
    )

    authtype = parser.add_mutually_exclusive_group()
    authtype.add_argument(
        "username",
        nargs="?",
        default=argparse.SUPPRESS,
        help="authentication username, required if --kerberos not given",
    )

    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        default=False,
        help="write debug output to stdout (implies --verbose)",
    )
    parser.add_argument(
        "-f",
        "--file",
        default=get_x509_proxy_path(),
        help="certificate file to create/reuse/destroy",
    )
    parser.add_argument(
        "-H",
        "--hours",
        type=int,
        default=277,
        help="lifetime of the certificate"
    )
    parser.add_identity_provider_argument()
    authtype.add_argument(
        "-k",
        "--kerberos",
        action="store_true",
        default=False,
        help="enable kerberos negotiation, required if username not given"
    )
    parser.add_argument(
        "-p",
        "--proxy",
        action="store_true",
        default=False,
        help="create RFC 3820 compliant impersonation proxy"
    )
    parser.add_argument(
        "-r",
        "--reuse",
        default=False,
        type=float,
        const=1.,
        nargs="?",
        metavar="HOURS",
        help="reuse an existing certificate if valid for more than "
             "%(const)s hours, or pass a number of hours to specify",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="write verbose output to stdout",
    )
    parser.add_argument(
        "-X",
        "--destroy",
        action="store_true",
        default=False,
        help="destroy existing certificate"
    )
    return parser


def parse_args(parser):
    """Parse and validate the command-line arguments

    Returns
    -------
    args : `argparse.Namespace`
    """
    args = parser.parse_args()

    # check that username or --kerberos was given if not using --destroy
    if not args.destroy and not (
            getattr(args, "username", None) or args.kerberos
    ):
        parser.error("one of username or -k/--kerberos is required")

    if args.debug:
        args.verbose = True

    return args


def can_reuse(path, proxy=None):
    try:
        check_cert(load_cert(path), proxy=proxy)
    except (RuntimeError, OSError):
        return False
    return True


def main():
    parser = create_parser()
    args = parse_args(parser)

    # if asked to destroy, just do that
    if args.destroy:
        if args.verbose:
            print("Removing credential file {!s}".format(args.file))
        os.unlink(args.file)
        sys.exit()

    # if asked to reuse, check that we can
    if args.reuse:
        if args.verbose:
            print("Validating existing certificate...", end=" ")
        args.reuse = can_reuse(args.file, proxy=args.proxy)
        if args.verbose and args.reuse:
            print("OK")
        elif args.verbose:
            print("failed, will regenerate")

    # get new certificate
    if not args.reuse:
        if args.verbose:
            print("Fetching certificate...")
        cert = get_cert(
            args.identity_provider,
            username=getattr(args, "username", None),
            kerberos=args.kerberos,
            hours=args.hours,
            debug=args.debug,
        )

        # write certificate to a file
        if args.verbose:
            print("Storing certificate...")
        write_cert(
            args.file,
            cert,
            use_proxy=args.proxy,
            minhours=args.hours,
        )
        if args.verbose:
            print("X.509 credential stored")

    # load the cert from file to print information
    if args.debug or args.verbose:
        x509 = load_cert(args.file)

    # dump full text of cert
    if args.debug:
        certstr = crypto.dump_certificate(
            crypto.FILETYPE_PEM,
            x509,
        )
        print(certstr.decode("utf-8"), end="")

    # print certificate/proxy info
    if args.verbose:
        print_cert_info(x509, path=args.file)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2019-2020)
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

import calendar
import datetime
import os
import math
import random
import string
import struct
import tempfile
import time
try:
    from urllib import parse as urllib_parse
    from urllib import request as urllib_request
except ImportError:  # python < 3
    import urlparse as urllib_parse
    import urllib2 as urllib_request

from OpenSSL import crypto

from M2Crypto import (X509, RSA, EVP, ASN1, m2)

from .ecp import (
    authenticate,
)
from .http import build_verified_opener

from .utils import (
    DEFAULT_SP_URL,
    format_endpoint_url,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- utilities ----------------------------------------------------------------

def _parse_unix_time(asn1time):
    return calendar.timegm(asn1time.get_datetime().timetuple())


def _random_string(length, outof=string.ascii_lowercase+string.digits):
    # http://stackoverflow.com/a/23728630/2213647 says SystemRandom()
    # is most secure
    return ''.join(random.SystemRandom().choice(outof) for _ in range(length))


# -- X.509 generator ----------------------------------------------------------

def get_x509_proxy_path():
    """Returns the default path for the X.509 certificate file

    Returns
    -------
    path : `str`
    """
    if os.getenv("X509_USER_PROXY"):
        return os.environ["X509_USER_PROXY"]
    if os.name == "nt":
        tmpdir = r'%SYSTEMROOT%\Temp'
        tmpname = "x509up_{0}".format(os.getlogin())
    else:
        tmpdir = "/tmp"
        tmpname = "x509up_u{0}".format(os.getuid())
    return os.path.join(tmpdir, tmpname)


def get_cert(
        endpoint,
        username=None,
        kerberos=False,
        hours=168,
        spurl=DEFAULT_SP_URL,
        debug=False,
):
    endpoint = format_endpoint_url(endpoint)

    # handle kerberos endpoint
    if kerberos is True:
        kerberos = endpoint

    # authenticate with IdP
    cookie, _ = authenticate(
        endpoint,
        username=username,
        kerberos=kerberos,
        debug=debug,
    )
    if debug:
        print("Authenticated against IdP with ECP")
        print("Requesting certificate from {0}".format(spurl))

    # request PKCS12 certificate from SP
    csrfstr = _random_string(10)
    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Cookie': 'CSRF={0}; {1.name}={1.value}'.format(csrfstr, cookie),
    }
    p12password = (
        _random_string(16, string.ascii_letters) +
        _random_string(2, string.digits) +
        _random_string(2, '!@#$%^&*()')
    )
    certdata = urllib_parse.urlencode({
        'submit': 'pkcs12',
        'CSRF': csrfstr,
        'p12password': p12password,
        'p12lifetime': math.ceil(hours),
    }).encode('utf-8')
    request = urllib_request.Request(
        url=spurl,
        headers=headers,
        data=certdata,
    )
    opener = build_verified_opener(debug=debug, krb_endpoint=kerberos)
    response = opener.open(request)

    if debug:
        print("Certificate received.")
    return crypto.load_pkcs12(
        response.read(),
        p12password.encode('utf-8'),
    )


# -- X.509 handling -----------------------------------------------------------

def load_cert(path, format=crypto.FILETYPE_PEM):
    """Load certificate from file

    Parameters
    ----------
    path : `str`
        the file path from which to read

    format : `int`, optional
        the format to read

    Returns
    -------
    cert : `OpenSSL.crypto.X509`
        the parsed certificate
    """
    with open(path, "r") as fobj:
        return crypto.load_certificate(
            format,
            fobj.read(),
        )


def _timeleft(cert):
    """Returns the number of seconds left on this certificate
    """
    try:  # M2Crypto
        expiry = cert.get_not_after().get_datetime().timetuple()
    except AttributeError:  # OpenSSL
        expiry = time.strptime(
            cert.get_notAfter().decode("utf-8"),
            "%Y%m%d%H%M%SZ",
        )
    return int(calendar.timegm(expiry) - time.time())


def _x509_name_str(obj):
    return "/" + "/".join([
        b"=".join(x).decode("utf-8") for x in obj.get_components()
    ])


def check_cert(cert, hours=1, proxy=None, rfc3820=True):
    """Validate an X509 certificate

    Parameters
    ----------
    cert : `OpenSSL.crypto.X509`
        the certificate object to check

    hours : `float`, optional
        minimum number of hours remaining before expiry

    proxy : `bool`, `NoneType`, optional
        if `True` (`False`), validate that the certificate is (is not)
        an impersonation proxy, if `None` (default) don't check.

    rfc3820 : `bool`, optional
        if `True` assert that, if the certificate is a proxy, that it
        is RFC 3820 compliant
    """
    # check expiry
    remaining = _timeleft(cert)
    if remaining < hours * 3600.:
        raise RuntimeError(
            "less than {0} hours remaining on X509 certificate".format(hours)
        )

    # check proxy type
    ctype = _cert_type(cert)
    isproxy = "proxy" in ctype
    if proxy is True and not isproxy:
        raise RuntimeError("certificate is not a proxy")
    if proxy is False and isproxy:
        raise RuntimeError("certificate is a proxy")
    if rfc3820 and "legacy globus proxy" in ctype:
        raise RuntimeError("proxy certificate is not RFC 3820 compliant")


def _cert_type(x509):
    # parse name entry as common name
    sub = x509.get_subject().get_components()
    if sub[-1][0] != b"CN":
        return "end entity credential"
    if sub[-1][1] == "proxy":
        return "full legacy globus proxy"
    if sub[-1][1] == "limited proxy":
        return "limited legacy globus proxy"

    # parse extensions for proxyCertInfo
    extensions = [x509.get_extension(i) for
                  i in range(x509.get_extension_count())]
    names = [e.get_short_name() for e in extensions]
    if b"proxyCertInfo" in names:
        return "RFC 3820 compliant impersonation proxy"
    return "end entity credential"


def print_cert_info(x509, path=None):
    """Print info about an X.509 certificate
    """
    pkey = x509.get_pubkey()
    print("subject  : " + _x509_name_str(x509.get_subject()))
    print("issuer   : " + _x509_name_str(x509.get_issuer()))
    print("type     : " + _cert_type(x509))
    print("strength : {0} bits".format(pkey.bits()))
    if path:
        print("path     : " + str(path))
    print("timeleft : " + str(datetime.timedelta(seconds=_timeleft(x509))))


def write_cert(path, pkcs12, use_proxy=False, minhours=168):
    """Write a PKCS12 certificate archive to file in X509 format

    Parameters
    ----------
    path : `str`
        the desired location of the final X509 file

    pkcs12 : `OpenSSL.crypto.PKCS12`
        a PKCS12 archive
    """
    certstr = crypto.dump_certificate(
        crypto.FILETYPE_PEM,
        pkcs12.get_certificate(),
    )
    keystr = crypto.dump_privatekey(
        crypto.FILETYPE_PEM,
        pkcs12.get_privatekey(),
    )
    if use_proxy:
        cert = X509.load_cert_string(certstr)
        key = EVP.load_key_string(keystr).get_rsa()
        proxy, proxykey = generate_proxy(cert, key, minhours=minhours)
        blocks = [proxy.as_pem(), proxykey.as_pem(cipher=None), certstr]
    else:
        blocks = [certstr, keystr]

    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        # write cert and key to temporary location
        for block in blocks:
            tmp.write(block)
        tmp.close()

        # move tmpfile into place
        os.rename(tmp.name, path)


def generate_proxy(cert, key, minhours=168, limited=False, bits=2048):
    """Generate a proxy certificate based on a certificate

    Based on code from the gridproxy library
    https://github.com/abbot/gridproxy/blob/master/gridproxy/__init__.py
    which is Copyright Lev Shamardin and covered under the GNU GPLv3 license.

    Returns
    -------
    proxycert : `M2Crypto.X509.X509`
        the proxy certificate

    proxykey : `M2Crypto.EVP.PKey`
        the proxy private key
    """
    # according to
    #   https://en.wikipedia.org/wiki/RSA_%28cryptosystem%29#Key_generation
    # the exponent 65537 (2^16+1) is most efficient
    proxyrsa = RSA.gen_key(bits, 65537, lambda x: None)
    proxykey = EVP.PKey()
    proxykey.assign_rsa(proxyrsa)

    proxy = X509.X509()
    proxy.set_pubkey(proxykey)
    proxy.set_version(2)

    not_before = ASN1.ASN1_UTCTIME()
    not_before.set_time(_parse_unix_time(cert.get_not_before()))
    proxy.set_not_before(not_before)
    now = int(time.time())
    not_after = ASN1.ASN1_UTCTIME()
    not_after_time = now + int(minhours * 60 * 60)
    # make sure proxy doesn't expire later than the underlying cert
    cert_not_after_time = _parse_unix_time(cert.get_not_after())
    if not_after_time > cert_not_after_time:
        not_after_time = cert_not_after_time
    not_after.set_time(not_after_time)
    proxy.set_not_after(not_after)

    proxy.set_issuer_name(cert.get_subject())
    digest = EVP.MessageDigest('sha256')
    digest.update(proxykey.as_der())
    serial = struct.unpack("<L", digest.final()[:4])[0]
    proxy.set_serial_number(int(serial & 0x7fffffff))

    # It is not completely clear what happens with memory allocation
    # within the next calls, so after building the whole thing we are
    # going to reload it through der encoding/decoding.
    proxy_subject = X509.X509_Name()
    subject = cert.get_subject()
    for idx in range(subject.entry_count()):
        entry = subject[idx].x509_name_entry
        m2.x509_name_add_entry(proxy_subject._ptr(), entry, -1, 0)
    proxy_subject.add_entry_by_txt('CN', ASN1.MBSTRING_ASC,
                                   str(serial), -1, -1, 0)
    proxy.set_subject(proxy_subject)
    proxy.add_ext(X509.new_extension(
        "keyUsage", "Digital Signature, Key Encipherment, Data Encipherment",
        1))
    if limited:
        proxy.add_ext(X509.new_extension(
            "proxyCertInfo", "critical, language:1.3.6.1.4.1.3536.1.1.1.9", 1))
    else:
        proxy.add_ext(X509.new_extension(
            "proxyCertInfo", "critical, language:Inherit all", 1))

    sign_pkey = EVP.PKey()
    sign_pkey.assign_rsa(key, 0)
    proxy.sign(sign_pkey, 'sha256')

    return proxy, proxykey



# -*- coding: utf-8 -*-
# Copyright (C) FNAL (2015-2016), Cardiff University (2019-2020)
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""HTTP methods for cigetcert
"""

import re
import socket
import warnings
try:
    from http import client as http_client
    from urllib import (
        request as urllib_request,
        parse as urllib_parse,
    )
    from urllib.error import HTTPError
except ImportError:  # python < 3
    import httplib as http_client
    import urllib2 as urllib_request
    import urlparse as urllib_parse
    URLError = urllib_request.HTTPError

try:
    import kerberos
except ImportError:  # windows
    import winkerberos as kerberos

from M2Crypto import (
    SSL,
    m2,
)

from .utils import (
    DEFAULT_CAFILE,
    DEFAULT_CAPATH,
)

__author__ = "Dave Dykstra, Duncan Macleod"


class _SslSocketWrapper(object):
    """Make a wrapper so a SSL Connection object can be opened as a file

    This is mostly from http://git.ganeti.org/?p=ganeti.git;a=commitdiff;h=beba56ae8;hp=70c815118f7f8bf151044cb09868d1e3d7a63ac8
    """  # noqa: E501
    def __init__(self, conn):
        self._conn = conn

    def __getattr__(self, name):
        # forward everything to underlying connection
        return getattr(self._conn, name)

    def makefile(self, mode, bufsize=0):
        try:
            return socket.SocketIO(self._conn, mode)
        except AttributeError:  # python < 3
            return socket._fileobject(self._conn, mode, bufsize)

    def close(self):
        # m2crypto always shuts down the SSL connection in the connection
        # close() function
        ret = self._conn.close()
        if (self._conn.get_shutdown() & 1) != 1:
            warnings.warn('close did not send shutdown')
        return ret


class CertValidatingHTTPSConnection(http_client.HTTPConnection):
    """Validate a certificate on an HTTPS connection with M2Crypto
    """
    default_port = http_client.HTTPS_PORT

    def __init__(self, host, port=None, key_file=None, cert_file=None,
                 cert_chain_file=None, cafile=None, capath=None, **kwargs):
        http_client.HTTPConnection.__init__(self, host, port, **kwargs)
        self.host = host
        self.key_file = key_file
        self.cert_file = cert_file
        self.cert_chain_file = cert_chain_file
        self.cafile = cafile
        self.capath = capath

    def connect(self):
        # 'sslv23' actually means to accept all, then we turn off
        #   insecure sslv2 & sslv3
        context = SSL.Context('sslv23')
        # many websites say to disable SSL compression for CRIME attack
        #  but unfortunately m2crypto doesn't have a constant for it
        SSL_OP_NO_COMPRESSION = 0x00020000
        context.set_options(m2.SSL_OP_NO_SSLv2 |
                            m2.SSL_OP_NO_SSLv3 | SSL_OP_NO_COMPRESSION)
        # the example used for M2Crypto connections was mostly
        #   https://www.heikkitoivonen.net/blog/2008/10/14/ssl-in-python-26/
        if (self.cert_file is not None) or (self.key_file is not None):
            context.load_cert(self.cert_file, self.key_file)
        if self.cert_chain_file is not None:
            context.load_cert_chain(self.cert_chain_file)
        # Note that m2crypto does not verify CRLs.  There is an
        # extension package m2ext that does, but ignoring CRLs for the
        # well-managed servers that cigetcert connects to is deemed to
        # be an acceptable risk.
        context.set_verify(
            SSL.verify_peer | SSL.verify_fail_if_no_peer_cert, depth=9)
        if context.load_verify_locations(self.cafile, self.capath) != 1:
            raise RuntimeError('Could not load verify locations ' +
                               str(self.cafile) + ' ' + str(self.capath))
        # The following cipher list is from "Disable weak ciphers" on
        # https://www.owasp.org/index.php/Transport_Layer_Protection_Cheat_Sheet
        # combined with the openssl man page at
        # https://www.openssl.org/docs/manmaster/apps/ciphers.html
        # You can see what remains by passing this list to
        #   openssl ciphers -v.
        if context.set_cipher_list('DEFAULT:!eNULL:!aNULL:!ADH:!EXP:!LOW:'
                                   '!MD5:!IDEA:!RC4:@STRENGTH') != 1:
            warnings.warn("no valid ciphers")
        sslconn = SSL.Connection(context)
        timeout = SSL.timeout(15)
        sslconn.set_socket_read_timeout(timeout)
        sslconn.set_socket_write_timeout(timeout)
        sslconn.connect((self.host, self.port))
        self.sock = _SslSocketWrapper(sslconn)


# -- handlers -----------------------------------------------------------------

class VerifiedHTTPSHandler(urllib_request.HTTPSHandler):
    def __init__(self, debuglevel=0, **kwargs):
        urllib_request.HTTPSHandler.__init__(self, debuglevel=debuglevel)
        self._connection_args = kwargs

    def http_class_wrapper(self, host, **kwargs):
        full_kwargs = dict(self._connection_args)
        full_kwargs.update(kwargs)
        return CertValidatingHTTPSConnection(host, **full_kwargs)

    def https_open(self, req):
        return self.do_open(self.http_class_wrapper, req)

    def http_error_401(self, request, response, code, msg, hdrs):
        return response


class NoRedirectHandler(urllib_request.HTTPRedirectHandler):
    def http_error_302(self, request, response, code, msg, hdrs):
        return response


class NoHttpHandler(urllib_request.HTTPHandler):
    def http_open(self, *args, **kwargs):
        raise ValueError('only https:// and file:// supported')


class HTTPNegotiateAuthHandler(urllib_request.BaseHandler):
    """Kerberos-based authentication handler

    This class uses an existing Kerberos ticket to authenticate
    via HTTP Negotiate Authentication. An instance of this class
    can be passed into the build_opener function from the urllib.request
    module.

    Modified from source found at

    http://selenic.com/pipermail/mercurial/2008-June/019776.html

    Parameters
    ----------
    service_principal : `str`, optional
        the Kerberos principal of the authentication host
    """
    rx = re.compile(r'(?:.*,)*\s*Negotiate\s*([^,]*),?', re.I)
    handler_order = 480  # before Digest auth

    def __init__(self, service_principal):
        """Create a new `HTTPNegotiateAuthHandler`
        """
        self.retried = 0
        self.context = None
        self.service_principal = service_principal

    def negotiate_value(self, headers):
        authreq = headers.get('www-authenticate', None)
        if authreq:
            mo = self.rx.search(authreq)
            if mo:
                return mo.group(1)
        return None

    def generate_request_header(self, req, headers):
        neg_value = self.negotiate_value(headers)
        if neg_value is None:
            self.retried = 0
            return None

        if self.retried > 5:
            raise HTTPError(
                req.get_full_url(), 401, "negotiate auth failed",
                headers, None)
        self.retried += 1

        result, self.context = kerberos.authGSSClientInit(
            self.service_principal,
        )
        if result < 1:
            return None

        if kerberos.authGSSClientStep(self.context, neg_value) < 0:
            return None
        return 'Negotiate {0}'.format(
            kerberos.authGSSClientResponse(self.context),
        )

    def authenticate_server(self, headers):
        neg_value = self.negotiate_value(headers)
        if neg_value is None:
            return None
        elif kerberos.authGSSClientStep(self.context, neg_value) < 1:
            pass

    def clean_context(self):
        if self.context is not None:
            kerberos.authGSSClientClean(self.context)

    def http_error_401(self, req, fp, code, msg, headers):
        try:
            neg_hdr = self.generate_request_header(req, headers)

            if neg_hdr is None:
                return None

            req.add_unredirected_header('Authorization', neg_hdr)
            resp = self.parent.open(req)

            self.authenticate_server(resp.info())

            return resp

        finally:
            self.clean_context()


# -- openers ------------------------------------------------------------------

def build_verified_opener(cookiejar=None, debug=False, cafile=DEFAULT_CAFILE,
                          capath=DEFAULT_CAPATH, redirect=False,
                          krb_endpoint=None):
    handlers = [
        urllib_request.HTTPCookieProcessor(cookiejar=cookiejar),
        VerifiedHTTPSHandler(cafile=cafile, capath=capath,
                             debuglevel=int(debug)),
        NoHttpHandler(),
    ]
    if not redirect:
        handlers.append(NoRedirectHandler())
    if krb_endpoint:
        loginhost = urllib_parse.urlparse(krb_endpoint).netloc.split(':')[0]
        handlers.append(
            HTTPNegotiateAuthHandler('HTTP@{0}'.format(loginhost))
        )
    return urllib_request.build_opener(*handlers)



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2019-2020)
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

r"""Authenticate and store SAML/ECP session cookies.

There are two usages:

    $ ecp-get-cookie -i 'My Institution' -u jsmith https://campus01.edu/my/secret/page

to authenticate with a password prompt, or

    $ ecp-get-cookie -i 'My Institution' -k https://campus01.edu/my/secret/page

to reuse an existing kerberos (``kinit``) credential.

By default the cookie file is created and stored in a location
defined by either

- ``/tmp/ecpcookie.u{uid}`` (Unix), or
- ``C:\Windows\Temp\ecpcookie.{username}`` (Windows)
"""  # noqa: E501

from __future__ import print_function

import argparse
import os
import sys

from ..cookies import (
    COOKIE_FILE as DEFAULT_COOKIE_FILE,
    ECPCookieJar,
)
from ..ecp import authenticate
from .utils import (
    ArgumentParser,
    reuse_cookiefile,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def create_parser():
    """Create a command-line argument parser

    Returns
    -------
    parser : `argparse.ArgumentParser`
    """
    parser = ArgumentParser(
        description=__doc__,
        prog="ecp-get-cookie",
    )
    parser.add_argument(
        "target_url",
        metavar="URL",
        help="service URL for which to generate cookies",
    )
    authtype = parser.add_mutually_exclusive_group()
    authtype.add_argument(
        "username",
        nargs="?",
        default=argparse.SUPPRESS,
        help="authentication username, required if --kerberos not given",
    )

    parser.add_argument(
        "-c",
        "--cookiefile",
        metavar="cookiefile",
        default=DEFAULT_COOKIE_FILE,
        help="cookie file to create/reuse/destroy",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        default=False,
        help="write debug output to stdout (implies --verbose)",
    )
    parser.add_identity_provider_argument()
    authtype.add_argument(
        "-k",
        "--kerberos",
        action="store_true",
        default=False,
        help="enable kerberos negotiation, required if username not given"
    )
    parser.add_argument(
        "-r",
        "--reuse",
        default=False,
        action="store_true",
        help="reuse existing cookies if possible",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="write verbose output to stdout",
    )
    parser.add_argument(
        "-X",
        "--destroy",
        action="store_true",
        default=False,
        help="destroy existing cookie file"
    )
    return parser


def parse_args(parser):
    """Parse and validate the command-line arguments

    Returns
    -------
    args : `argparse.Namespace`
    """
    args = parser.parse_args()

    # check that username or --kerberos was given if not using --destroy
    if not args.destroy and not (
            args.kerberos or
            getattr(args, "username", None) and args.idp
    ):
        parser.error("-k/--kerberos is required if IdP_tag and "
                     "login are not given")

    if args.debug:
        args.verbose = True

    return args


def main():
    parser = create_parser()
    args = parse_args(parser)

    # if asked to destroy, just do that
    if args.destroy:
        if args.verbose:
            print("Removing cookie file {0!s}".format(args.cookiefile))
        os.unlink(args.cookiefile)
        sys.exit()

    # if asked to reuse, check that we can
    cookiejar = None
    if args.reuse:
        cookiejar, args.reuse = reuse_cookiefile(
            args.cookiefile,
            args.target_url,
            verbose=args.verbose,
        )

    # get new certificate
    if not args.reuse:
        cookiejar = cookiejar or ECPCookieJar()
        if args.verbose:
            print("Authenticating...")
        authenticate(
            args.identity_provider,
            spurl=args.target_url,
            cookiejar=cookiejar,
            username=getattr(args, "username", None),
            kerberos=args.kerberos,
            debug=args.debug,
        )

        # write certificate to a file
        if args.verbose:
            print("Storing cookies...")
        cookiejar.save(
            args.cookiefile,
            ignore_discard=True,
            ignore_expires=True,
        )
        if args.verbose:
            print("Cookies stored in '{0!s}'".format(args.cookiefile))

    # load the cert from file to print information
    if args.debug or args.verbose:
        info = [(cookie.domain, cookie.path, cookie.secure, cookie.name) for
                cookie in cookiejar]
        fmt = "%-{0}s %-5s %-7s %s".format(max(len(i[0]) for i in info))
        print(fmt % ("------", "----", "-------", "----"))
        print(fmt % ("DOMAIN", "PATH", r"SECURE?", "NAME"))
        print(fmt % ("------", "----", "-------", "----"))
        for tup in info:
            print(fmt % tup)
        print(fmt % ("------", "----", "-------", "----"))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2019-2020)
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Common utilities for tools
"""

from __future__ import print_function

import argparse
import sys

from .. import __version__
from ..cookies import (
    LoadError,
    has_session_cookies,
    load_cookiejar,
)
from ..env import _get_default_idp
from ..utils import get_idps

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


# -- argparse helpers ---------------------------

class HelpFormatter(
    argparse.ArgumentDefaultsHelpFormatter,
    argparse.RawDescriptionHelpFormatter,
):
    def _format_usage(self, usage, actions, groups, prefix):
        if prefix is None:
            prefix = "Usage: "
        return super(HelpFormatter, self)._format_usage(
            usage,
            actions,
            groups,
            prefix,
        )


class ArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        add_helpers = kwargs.pop("add_helpers", True)
        version = kwargs.pop("version", __version__)
        manpage = kwargs.pop("manpage", None)

        kwargs.setdefault("formatter_class", HelpFormatter)
        kwargs.setdefault("add_help", not add_helpers)
        super(ArgumentParser, self).__init__(*args, **kwargs)

        self._positionals.title = "Required arguments"
        self._optionals.title = "Options"

        # add manpage options for argparse-manpage
        self._manpage = manpage

        # add helper commands group
        if add_helpers:
            helpers = self.add_argument_group("Helper arguments")
            helpers.add_argument(
                "-h",
                "--help",
                action="help",
                default=argparse.SUPPRESS,
                help="show this help message and exit",
            )
            helpers.add_argument(
                "-l",
                "--list-idps",
                action=ListIdpsAction,
            )
            if version is not None:
                helpers.add_argument(
                    "-V",
                    "--version",
                    action="version",
                    version=version,
                )

    def add_identity_provider_argument(
            self,
            *args,
            **kwargs
    ):
        if not args:
            args = ("-i", "--identity-provider")
        kwargs.setdefault("default", _get_default_idp())
        kwargs.setdefault(
            "help",
            "name of institution providing the identity (e.g. 'Cardiff "
            "University'), or domain name of IdP host (e.g. idp.cf.ac.uk), "
            "see --list-idps for a list of Identity Provider (IdPs) and "
            "their IdP URL. "
            "Shortened institution names (e.g. 'Cardiff') can be given as "
            "long as they uniquely match a full institution name known by "
            "CILogon",
        )
        return super(ArgumentParser, self).add_argument(
            *args,
            **kwargs
        )


class ListIdpsAction(argparse.Action):
    def __init__(
            self,
            option_strings,
            dest=argparse.SUPPRESS,
            default=argparse.SUPPRESS,
            help="show IdP names and URLs and exit",
    ):
        super(ListIdpsAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        idps = get_idps()
        formatter = parser._get_formatter()
        fmt = "%-{0}r : %s".format(max(map(len, idps)) + 2)
        lines = []
        for pair in sorted(idps.items(), key=lambda x: x[0]):
            lines.append(fmt % pair)
        formatter.add_text('\n'.join(lines))
        parser._print_message(formatter.format_help(), sys.stdout)
        parser.exit()


# -- miscellaneous ------------------------------

def reuse_cookiefile(cookiefile, url, verbose=False):
    """Load cookies from a cookiefile and determine if they can be reused

    Parameters
    ----------
    cookiefile : `str`
        the path to the cookie file

    url : `str`
        the URL of the service being accessed with cookies

    verbose : `bool`, optional
        if `True`, print verbose output

    Returns
    -------
    cookiejar : `ligo.org.cookies.ECPCookieJar`, `NoneType`
        the cookie jar (if the cookie file was read correctly), otherwise
        `None`
    reuse : `bool`
        `True` if the cookies in the jar can be reused to access the given
        service URL, otherwise `False`
    """
    if verbose:
        print("Validating existing cookies...", end=" ")
    try:
        cookiejar = load_cookiejar(cookiefile, strict=True)
    except (LoadError, OSError):
        if verbose:
            print("failed to load cookie file")
        return None, None
    reuse = has_session_cookies(cookiejar, url)
    if verbose and reuse:
        print("OK")
    elif verbose:
        print("cannot reuse, need new cookies")
    return cookiejar, reuse



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2019-2020)
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""ECP authentication
"""

from __future__ import (print_function, absolute_import)

import base64
try:
    from urllib import (
        request as urllib_request,
        parse as urllib_parse,
    )
    from urllib.error import URLError
    from http.cookiejar import CookieJar
except ImportError:  # python < 3
    import urllib2 as urllib_request
    import urlparse as urllib_parse
    from cookielib import CookieJar
    URLError = urllib_request.URLError

from lxml import etree

from .http import build_verified_opener
from .utils import (
    DEFAULT_SP_URL,
    format_endpoint_url,
    prompt_username_password,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def get_xml_attribute(xdata, path, namespaces=None):
    if namespaces is None:
        namespaces = {
            'ecp': 'urn:oasis:names:tc:SAML:2.0:profiles:SSO:ecp',
            'S': 'http://schemas.xmlsoap.org/soap/envelope/',
            'paos': 'urn:liberty:paos:2003-08'
        }
    return xdata.xpath(path, namespaces=namespaces)[0]


def report_soap_fault(opener, url):
    soapfault = """
        <S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/">
           <S:Body>
             <S:Fault>
                <faultcode>S:Server</faultcode>
                <faultstring>responseConsumerURL from SP and assertionConsumerServiceURL from IdP do not match</faultstring>
             </S:Fault>
           </S:Body>
        </S:Envelope>"""  # noqa
    headers = {'Content-Type': 'application/vnd.paos+xml'}
    request = urllib_request.Request(url, headers=headers, data=soapfault)
    return opener.open(request)


def authenticate(
        endpoint,
        kerberos=False,
        spurl=DEFAULT_SP_URL,
        cookiejar=None,
        username=None,
        debug=False,
        **kwargs
):
    """Authenticate against an endpoint using ECP

    Parameters
    ----------
    endpoint : `str`
        the URL of the ECP endpoint to negotiate with

    kerberos : `bool`, or `str`, optional
        a `str` denoting the endpoint to use for kerberos negotiation,
        or `True` to use kerberos negotiation with the standard endpoint,
        otherwise do not use kerberos

    username : `str`, optional
        the username registered at the endpoint, if not given, and
        not using kerberos, this will be prompted for

    spurl : `str`, optional
        the URL of the service provider to authenticate against

    cookiejar : `http.cookiejar.CookieJar`, optional
        a cookie jar to store cookies in

    **kwargs
        other keyword arguments are passed to
        :func:`ligo.org.utils.build_verified_opener`

    Returns
    -------
    cookie : `http.cookiejar.Cookie`
        the newly created shibsession cookie
    cookiejar : `http.cookiejar.CookieJar`
        the cookiejar that contains the cookie(s)
    """
    endpoint = format_endpoint_url(endpoint)

    if kerberos is True:
        kerberos = endpoint

    # build HTTP request opener (if not given)
    if cookiejar is None:
        cookiejar = CookieJar()
    opener = build_verified_opener(
        cookiejar=cookiejar,
        krb_endpoint=kerberos,
        debug=debug,
        **kwargs
    )

    # -- step 1: initiate ECP request -----------

    # headers needed to indicate to the SP an ECP request
    headers = {
        'Accept': 'text/html; application/vnd.paos+xml',
        'PAOS': 'ver="urn:liberty:paos:2003-08";'
                '"urn:oasis:names:tc:SAML:2.0:profiles:SSO:ecp"',
    }

    # request target from SP
    request = urllib_request.Request(
        url=spurl or endpoint,
        headers=headers,
    )
    response = opener.open(request)
    data = response.read()
    if debug:
        print("##### begin SP response\n")
        print(data.decode("utf-8"))
        print("\n##### end SP response")

    # convert the SP resonse from string to etree Element object
    spetree = etree.XML(data)

    # pick out the relay state element from the SP so that it can
    # be included later in the response to the SP
    relaystate = get_xml_attribute(
        spetree,
        "//ecp:RelayState",
    )
    rcurl = get_xml_attribute(
        spetree,
        "/S:Envelope/S:Header/paos:Request/@responseConsumerURL",
    )

    # remote the SOAP header to create a packge for the IdP
    idpbody = spetree
    idpbody.remove(idpbody[0])

    # -- step 2: authenticate with endpoint -----

    request = urllib_request.Request(
        endpoint,
        method='POST',
        data=etree.tostring(idpbody),
    )
    request.add_header('Content-Type', 'text/xml; charset=utf-8')

    # get credentials for non-kerberos request
    if not kerberos:
        # prompt the user for a password
        username, password = prompt_username_password(
            urllib_parse.urlparse(endpoint).netloc.split(':')[0],
            username=username,
        )
        # using the Authorization header
        pair = '%s:%s' % (username, password)
        try:
            base64string = base64.encodebytes(
                pair.encode('utf-8')).decode('utf-8')
        except AttributeError:  # python < 3
            base64string = base64.encodestring(pair)
        request.add_header('Authorization', 'Basic {0}'.format(
            base64string.replace('\n', '')))

    response = opener.open(request)
    data = response.read()

    if debug:
        print("##### begin IdP response")
        print(data.decode("utf-8"))
        print("##### end IdPesponse")

    # -- step 3: post back to the SP ------------

    try:
        idptree = etree.XML(data)
    except etree.XMLSyntaxError:
        raise RuntimeError(
            "Failed to parse response from {}, you most "
            "likely incorrectly entered your passphrase".format(
                endpoint,
            ),
        )
    acsurl = get_xml_attribute(
        idptree,
        "/S:Envelope/S:Header/ecp:Response/@AssertionConsumerServiceURL")

    # validate URLs between SP and IdP
    if acsurl != rcurl:
        try:
            report_soap_fault(opener, rcurl)
        except URLError:
            pass  # don't care, just doing a service

    # make a deep copy of the IdP response and replace its
    # header contents with the relay state initially sent by
    # the SP
    actree = idptree
    actree[0][0] = relaystate

    # POST the package to the SP
    headers = {'Content-Type': 'application/vnd.paos+xml'}
    request = urllib_request.Request(url=acsurl, method='POST',
                                     headers=headers,
                                     data=etree.tostring(actree))
    response = opener.open(request)

    # -- done -----------------------------------
    # we should now have an ECP cookie in the cookie jar

    cookie = cookiejar.make_cookies(response, request)[0]
    return cookie, cookiejar



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2019-2020)
#
# This file is part of ciecplib
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Build configuration for ciecplib
"""

import os.path
import re

from setuptools import (find_packages, setup)
from setuptools.command.build_py import build_py
from setuptools.command.install import install

try:
    from build_manpages.build_manpages import (
        build_manpages,
        get_build_py_cmd,
        get_install_cmd,
    )
except ImportError:  # can't build manpages, that's ok
    cmdclass = {
        "build_py": build_py,
        "install": install,
    }
else:
    cmdclass = {
        "build_manpages": build_manpages,
        "build_py": get_build_py_cmd(build_py),
        "install": get_install_cmd(install),
    }

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def find_version(path, varname="__version__"):
    """Parse the version metadata in the given file.
    """
    with open(path, 'r') as fobj:
        version_file = fobj.read()
    version_match = re.search(
        r"^{0} = ['\"]([^'\"]*)['\"]".format(varname),
        version_file,
        re.M,
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup_requires = [
    "setuptools",  # MIT
]
install_requires = [
    "lxml",  # BSD
    "M2Crypto",  # MIT
    "pykerberos ; sys_platform != 'win32'",  # Apache 2.0
    "pyOpenSSL",  # Apache 2.0
    "winkerberos ; sys_platform == 'win32'",  # Apache 2.0
]
tests_require = [
    "mock ; python_version < '3'",
    "pytest",
    "pytest-cov",
]
extras_require = {
    "test": tests_require,
    "docs": [
        "sphinx",
        "sphinx-argparse",
        "sphinx_automodapi",
        "sphinx_rtd_theme",
        "sphinx_tabs",
    ],
}

setup(
    # distribution metadata
    name="ciecplib",
    version=find_version(os.path.join("ciecplib", "__init__.py")),
    author="Duncan Macleod",
    author_email="duncan.macleod@ligo.org",
    license="GPL-3.0-or-later",
    description="A python client for SAML ECP authentication",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    url="https://pypi.org/project/ciecplib/",
    project_urls={
        "Bug Tracker": "https://github.com/duncanmmacleod/ciecplib/issues",
        "Documentation": "https://ciecplib.readthedocs.io/",
        "Source Code": "https://github.com/duncanmmacleod/ciecplib/",
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering',
    ],
    # contents
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "ecp-cert-info=ciecplib.tool.ecp_cert_info:main",
            "ecp-curl=ciecplib.tool.ecp_curl:main",
            "ecp-get-cert=ciecplib.tool.ecp_get_cert:main",
            "ecp-get-cookie=ciecplib.tool.ecp_get_cookie:main",
        ],
    },
    # dependencies
    cmdclass=cmdclass,
    setup_requires=setup_requires,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,
)



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2019-2020)
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Utility module to initialise a kerberos ticket

This module provides a lazy python version of the 'kinit'
command-line tool, with internal guesswork using keytabs

See the documentation of the `kinit` function for example usage
"""

import getpass
import os
import re
import subprocess
import sys
from collections import OrderedDict
from distutils.spawn import find_executable

if sys.version_info.major < 3:
    input = raw_input  # noqa: F821

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = ['kinit']

try:
    _IPYTHON = __IPYTHON__
except NameError:
    _IPYTHON = False


def _find_executable(name):
    exe = find_executable(name)
    if exe is None:
        raise OSError("cannot find {0!r}".format(name))
    return exe


class KerberosError(RuntimeError):
    """Kerberos (krb5) operation failed
    """
    pass


def kinit(username=None, password=None, realm=None, exe=None, keytab=None,
          krb5ccname=None, verbose=None):
    """Initialise a kerberos (krb5) ticket.

    This allows authenticated connections to, amongst others, NDS2
    services.

    Parameters
    ----------
    username : `str`, optional
        name of user, will be prompted for if not given.

    password : `str`, optional
        cleartext password of user for given realm, will be prompted for
        if not given.

    realm : `str`, optional
        name of realm to authenticate against, read from keytab if available

    exe : `str`, optional
        path to kinit executable.

    keytab : `str`, optional
        path to keytab file. If not given this will be read from the
        ``KRB5_KTNAME`` environment variable. See notes for more details.

    krb5ccname : `str`, optional
        path to Kerberos credentials cache.

    verbose : `bool`, optional
        print verbose output (if `True`), or not (`False`); default is `True`
        if any user-prompting is needed, otherwise `False`.

    Notes
    -----
    If a keytab is given, or is read from the ``KRB5_KTNAME`` environment
    variable, this will be used to guess the username and realm, if it
    contains only a single credential.

    Examples
    --------
    Example 1: standard user input, with password prompt::

       >>> kinit('albert.einstein', realm="LIGO.ORG")
       Password for albert.einstein@LIGO.ORG:
       Kerberos ticket generated for albert.einstein@LIGO.ORG

    Example 2: extract username and realm from keytab, and use that
    in authentication::

       >>> kinit(keytab='~/.kerberos/ligo.org.keytab', verbose=True)
       Kerberos ticket generated for albert.einstein@LIGO.ORG
    """
    # get kinit path
    exe = exe or _find_executable('kinit')

    # get keytab
    if keytab is None:
        keytab = os.environ.get('KRB5_KTNAME', None)
        if keytab is None or not os.path.isfile(keytab):
            keytab = None
    if keytab:
        try:
            principals = parse_keytab(keytab)
        except KerberosError:
            pass
        else:
            # is there's only one entry in the keytab, use that
            if username is None and len(principals) == 1:
                username = principals[0][0]
            # or if the given username is in the keytab, find the realm
            if username in list(zip(*principals))[0]:
                idx = list(zip(*principals))[0].index(username)
                realm = principals[idx][1]
            # otherwise this keytab is useless, so remove it
            else:
                keytab = None

    # refuse to prompt if we can't get an answer
    # note: jupyter streams are not recognised as interactive
    #       (isatty() returns False) so we have a special case here
    if not sys.stdout.isatty() and not _IPYTHON and (
            username is None or
            (not keytab and password is None)
    ):
        raise KerberosError("cannot generate kerberos ticket in a "
                            "non-interactive session, please manually create "
                            "a ticket, or consider using a keytab file")

    # get credentials
    if username is None and realm is None:
        verbose = True
        username, realm = input(
            "Please provide username, including Kerberos realm: ",
        ).split("@", 1)
    if username is None:
        verbose = True
        username = input("Please provide username for the {0} kerberos "
                         "realm: ".format(realm))
    identity = "{0}@{1}".format(username, realm)
    if not keytab and password is None:
        verbose = True
        password = getpass.getpass(
            prompt="Password for {0}: ".format(identity),
        )

    # format kinit command
    if keytab:
        cmd = [exe, "-k", "-t", keytab, identity]
    else:
        cmd = [exe, identity]
    if krb5ccname:
        krbenv = {"KRB5CCNAME": krb5ccname}
    else:
        krbenv = None

    # execute command
    kget = subprocess.Popen(cmd, env=krbenv, stdout=subprocess.PIPE,
                            stdin=subprocess.PIPE)
    if not keytab:
        kget.communicate(password.encode("utf-8"))
    kget.wait()
    retcode = kget.poll()
    if retcode:
        raise subprocess.CalledProcessError(kget.returncode, " ".join(cmd))
    if verbose:
        print("Kerberos ticket generated for {0}".format(identity))


def parse_keytab(keytab):
    """Read the contents of a KRB5 keytab file, returning a list of
    credentials listed within

    Parameters
    ----------
    keytab : `str`
        path to keytab file

    Returns
    -------
    creds : `list` of `tuple`
        the (unique) list of `(username, realm, kvno)` as read from the
        keytab file

    Examples
    --------
    >>> from ligo.org.kerberos import parse_keytab
    >>> print(parse_keytab("creds.keytab"))
    [('albert.einstein', 'LIGO.ORG', 1)]
    """
    try:
        out = subprocess.check_output(["klist", '-k', keytab],
                                      stderr=subprocess.PIPE)
    except OSError:
        raise KerberosError("Failed to locate klist, cannot read keytab")
    except subprocess.CalledProcessError:
        raise KerberosError("Cannot read keytab {0!r}".format(keytab))
    principals = []
    for line in out.splitlines():
        if isinstance(line, bytes):
            line = line.decode('utf-8')
        try:
            kvno, principal, = re.split(r'\s+', line.strip(' '), 1)
        except ValueError:
            continue
        else:
            if not kvno.isdigit():
                continue
            principals.append(tuple(principal.split('@')) + (int(kvno),))
    # return unique, ordered list
    return list(OrderedDict.fromkeys(principals).keys())


def has_credential(klist=None):
    """Run ``klist`` to determine whether a kerberos credential is available

    Parameters
    ----------
    `bool`
        `True` if ``klist -s`` indicates a valid credential
    """
    # run klist to get list of principals
    klist = klist or find_executable("klist")
    try:
        subprocess.check_output([klist, "-s"])
    except subprocess.CalledProcessError:
        return False
    return True



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2019-2020)
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

import os.path
import platform
import random
import re
import ssl
import string
from getpass import getpass
try:
    from urllib import request as urllib_request
    from urllib.parse import urlparse
except ImportError:  # python < 3
    import urllib2 as urllib_request
    from urlparse import urlparse
    input = raw_input  # noqa: F821

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# -- default CA files ---------------------------------------------------------

DEFAULT_VERIFY_PATHS = ssl.get_default_verify_paths()


def _defaults():
    if platform.system() in {"Windows", "Darwin"}:
        return DEFAULT_VERIFY_PATHS.cafile, DEFAULT_VERIFY_PATHS.capath

    # -- on Linux we can be a bit more discerning

    # always prefer python's default cafile
    for path in (
        DEFAULT_VERIFY_PATHS.cafile,
        "/etc/ssl/certs/ca-certificates.crt",  # debian
        "/etc/pki/tls/cert.pem",
    ):
        if os.path.isfile(path or ""):
            cafile = path
            break
    else:
        cafile = None

    # prefer globus certificates path
    globusdir = "/etc/grid-security/certificates"
    if os.path.isdir(globusdir):
        return cafile, globusdir

    return cafile, DEFAULT_VERIFY_PATHS.capath


DEFAULT_CAFILE, DEFAULT_CAPATH = _defaults()


# -- institution URLs ----------------------------------------------------

DEFAULT_IDPLIST_URL = "https://cilogon.org/include/ecpidps.txt"
DEFAULT_SP_URL = "https://ecp.cilogon.org/secure/getcert"
KERBEROS_SUFFIX = " (Kerberos)"
KERBEROS_REGEX = re.compile(r"{0}\Z".format(re.escape(KERBEROS_SUFFIX)))


def get_idps(url=DEFAULT_IDPLIST_URL):
    """Download the list of known ECP IdPs from the given URL

    The output is a `dict` where the keys are institution names
    (e.g. ``'Fermi National Accelerator Laboratory'``), and the values
    are the URL of their IdP.

    Some institutions may have two entries if they also support Kerberos.

    Parameters
    ----------
    url : `str`
        the URL of the IDP list file
    """
    idps = dict()
    for line in urllib_request.urlopen(url):
        url, inst = line.decode('utf-8').strip().split(' ', 1)
        idps[inst] = url
    return idps


def _match_institution(value, institutions):
    regex = re.compile(r"{0}($| \()".format(value))
    institutions = {KERBEROS_REGEX.split(name, 1)[0] for name in institutions}
    matches = [inst for inst in institutions if regex.match(inst)]
    if len(matches) == 1:
        return matches[0]
    if len(matches) == 0 and not value.endswith(".*"):
        try:
            return _match_institution("{0}.*".format(value), institutions)
        except ValueError:
            pass
    raise ValueError("failed to identify IdP URLs for {0!r}".format(value))


def get_idp_urls(institution, url=DEFAULT_IDPLIST_URL):
    """Return the regular and Kerberos IdP URLs for a given institution
    """
    idps = get_idps(url=url)
    institution = _match_institution(institution, idps)
    if institution.endswith(KERBEROS_SUFFIX):
        return None, idps[institution]
    url = idps[institution]
    krbinst = institution + KERBEROS_SUFFIX
    krburl = idps[krbinst] if krbinst in idps else url
    return url, krburl


def _endpoint_url(url):
    parsed = urlparse(url)
    if not parsed.scheme:
        url = "https://{0}".format(url)
    if not urlparse(url).path:
        return "{0}/idp/profile/SAML2/SOAP/ECP".format(url)
    return url


def format_endpoint_url(url_or_name, kerberos=False):
    """Format an endpoint reference as a URL

    Parameters
    ----------
    url_or_name: `str`
        the name of an institution, or a URL for the endpoint

    kerberos : `bool`, optional
        if ``True`` return a Kerberos URL, if available, otherwise return
        a standard SAML/ECP endpoint URL

    Returns
    -------
    url : `str`
        the formatted URL of the IdP ECP endpoint

    Raises
    ------
    ValueError
        if ``url_or_name`` looks like an institution name, but
        cilogon.org doesn't know what the corresponding ECP endpoint URL
        is for that institution.

    Examples
    --------
    >>> format_endpoint_url("LIGO")
    'https://login.ligo.org/idp/profile/SAML2/SOAP/ECP'
    >>> format_endpoint_url("login.myidp.com")
    'https://login.myidp.com/idp/profile/SAML2/SOAP/ECP'
    """
    if url_or_name.count(".") >= 2:  # url
        return _endpoint_url(url_or_name)
    # institution name
    return get_idp_urls(url_or_name)[int(kerberos)]


# -- misc utilities -----------------------------------------------------------

def random_string(length, outof=string.ascii_lowercase+string.digits):
    # http://stackoverflow.com/a/23728630/2213647 says SystemRandom()
    # is most secure
    return ''.join(random.SystemRandom().choice(outof) for _ in range(length))


def prompt_username_password(host, username=None):
    if username is None:
        username = input("Enter username for {0}: ".format(host))
    password = getpass(
        "Enter password for {0!r} on {1}: ".format(username, host),
    )
    return username, password



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2019-2020)
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Transfer a URL using SAML/ECP authentication.

This script attemps to use existing cookies for the target domain
and prompts for authorisation information if required.
Persistent cookies acquired during transactions are automatically
recorded in the cookie file, but session cookies are discarded (unless
--store-session-cookies is given).
"""

from __future__ import print_function

import sys

from .. import request
from ..cookies import COOKIE_FILE as DEFAULT_COOKIE_FILE
from .utils import (
    ArgumentParser,
    reuse_cookiefile,
)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def create_parser():
    """Create a command-line argument parser

    Returns
    -------
    parser : `argparse.ArgumentParser`
    """
    parser = ArgumentParser(
        description=__doc__,
        prog="ecp-curl",
    )
    parser.add_argument(
        "url",
        help="the URL to transfer"
    )
    parser.add_argument(
        "-c",
        "--cookiefile",
        metavar="FILE",
        default=DEFAULT_COOKIE_FILE,
        help="cookie file to use",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        default=False,
        help="write debug output to stdout",
    )
    parser.add_identity_provider_argument()
    parser.add_argument(
        "-k",
        "--kerberos",
        action="store_true",
        default=False,
        help="enable kerberos negotiation"
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        help="write to %(metavar)s instead of stdout",
    )
    parser.add_argument(
        "-s",
        "--store-session-cookies",
        action="store_true",
        default=False,
        help="store session cookies in the cookie file"
    )
    parser.add_argument(
        "-u",
        "--username",
        help="authentication username, will be prompted for if not given "
             "and not using --kerberos"
    )
    return parser


def main(args=None):
    parser = create_parser()
    args = parser.parse_args(args=args)
    cookiejar = reuse_cookiefile(
        args.cookiefile,
        args.url,
        verbose=False,
    )[0]
    if args.output:
        stream = open(args.output, "w")
    else:
        stream = sys.stdout
    try:
        print(
            request(
                args.url,
                cookiefile=args.cookiefile,
                cookiejar=cookiejar,
                endpoint=args.identity_provider,
                debug=args.debug,
                username=args.username,
                kerberos=args.kerberos,
                store_session_cookies=args.store_session_cookies,
            ).read().decode("utf-8"),
            file=stream,
            end="",
        )
    finally:
        if args.output:
            stream.close()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2019-2020)
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

r"""Print information about an existing X.509 credential.

"""  # noqa: E501

from __future__ import print_function

from ..x509 import (
    get_x509_proxy_path,
    load_cert,
    print_cert_info,
)
from .utils import (
    ArgumentParser,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

EPILOG = r"""
Environment:

X509_USER_PROXY:
    the default path for the credential file
"""

MANPAGE = [
    {'heading': 'environment',
     'content': """
.TP
.B "X509_USER_PROXY"
The default path for the credential file """,
     },
]


def create_parser():
    """Create a command-line argument parser

    Returns
    -------
    parser : `argparse.ArgumentParser`
    """
    parser = ArgumentParser(
        description=__doc__,
        epilog=EPILOG,
        manpage=MANPAGE,
        prog="ecp-cert-info",
    )
    parser.add_argument(
        "-f",
        "--file",
        default=get_x509_proxy_path(),
        help="certificate file to create/reuse/destroy",
    )
    return parser


def parse_args(parser):
    """Parse and validate the command-line arguments

    Returns
    -------
    args : `argparse.Namespace`
    """
    return parser.parse_args()


def main():
    parser = create_parser()
    args = parse_args(parser)
    print_cert_info(load_cert(args.file), path=args.file)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2019-2020)
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Cookie handling for SAML ECP authentication
"""

from __future__ import print_function

import os
import time
import warnings
try:
    from http.cookiejar import (LoadError, MozillaCookieJar)
    from urllib.parse import urlparse
except ImportError:  # python < 3
    from cookielib import (LoadError, MozillaCookieJar)
    from urlparse import urlparse

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def get_ecpcookie_path():
    if os.name == "nt":
        tmpdir = r'%SYSTEMROOT%\Temp'
        tmpname = "ecpcookie.{0}".format(os.getlogin())
    else:
        tmpdir = "/tmp"
        tmpname = "ecpcookie.u{0}".format(os.getuid())
    return os.path.join(tmpdir, tmpname)


COOKIE_FILE = get_ecpcookie_path()


# -- cookie jar ---------------------------------------------------------------

class ECPCookieJar(MozillaCookieJar):
    """Custom cookie jar

    Adapted from
    https://wiki.shibboleth.net/confluence/download/attachments/4358416/ecp.py
    """
    def save(self, filename=None, ignore_discard=False, ignore_expires=False):
        with open(filename, 'w') as f:
            f.write(self.header)
            for cookie in self:
                if not ignore_discard and cookie.discard:
                    continue
                if not ignore_expires and cookie.is_expired(time.time()):
                    continue
                if cookie.expires is not None:
                    expires = str(cookie.expires)
                else:
                    # change so that if a cookie does not have an expiration
                    # date set it is saved with a '0' in that field instead
                    # of a blank space so that the curl libraries can
                    # read in and use the cookie
                    expires = "0"
                if cookie.value is None:
                    # cookies.txt regards 'Set-Cookie: foo' as a cookie
                    # with no name, whereas cookiejar regards it as a
                    # cookie with no value.
                    name = ""
                    value = cookie.name
                else:
                    name = cookie.name
                    value = cookie.value
                print('\t'.join([
                    cookie.domain,
                    str(cookie.domain.startswith('.')).upper(),
                    cookie.path,
                    str(cookie.secure).upper(),
                    expires,
                    name,
                    value,
                ]), file=f)

    def _really_load(self, f, filename, ignore_discard, ignore_expires):
        out = super(ECPCookieJar, self)._really_load(
            f,
            filename,
            ignore_discard,
            ignore_expires,
        )
        for cookie in self:
            if not cookie.expires:
                cookie.expires = None
        return out


def has_session_cookies(jar, url):
    """Returns `True` if this ``jar`` contains session cookies we can reuse

    Parameters
    ----------
    jar : `http.cookiejar.CookieJar`
        the cookie jar to check

    url : `str`
        the URL of the service that needs cookies

    Returns
    -------
    can_reuse : `bool`
        `True` if any cookie in the jar is a non-expiring ``shibsession``
        cookie for the given same domain as ``url``
    """
    url = urlparse(url).netloc
    for cookie in jar:
        if (
                cookie.name.startswith("_shibsession_") and
                cookie.domain == url and
                cookie.expires is None
        ):
            return True
    return False


def load_cookiejar(
        cookiefile,
        strict=True,
        ignore_discard=True,
        ignore_expires=True,
):
    """Load a cookie jar from a file

    Parameters
    ----------
    cookiefile : `str`
        path to cookie jar file

    strict : `bool`, optional
        if `True` (default), raise all exceptions as they occur,
        if `False` just emit warnings

    ignore_discard, ignore_expires : `bool`, optional
        options to pass to :meth:`ECPCookieJar.load`, both default to
        `True` in this usage.
    """
    cookiejar = ECPCookieJar()
    try:
        cookiejar.load(
            str(cookiefile),
            ignore_discard=ignore_discard,
            ignore_expires=ignore_expires,
        )
    except (LoadError, OSError) as e:
        if strict:
            raise
        warnings.warn('Caught error loading ECP cookie: %s' % str(e))
    return cookiejar



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2019-2020)
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`ciecplib.utils`
"""

try:
    from unittest import mock
except ImportError:  # python < 3
    import mock
    urlreqmodname = "urllib2"
else:
    urlreqmodname = "urllib.request"

import pytest

from .. import utils as ciecplib_utils

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

RAW_INSTITUTION_LIST = [
    b"https://inst1.test Institution 1",
    b"https://inst2.test Institution 2",
]


@mock.patch(
    "{0}.urlopen".format(urlreqmodname),
    return_value=RAW_INSTITUTION_LIST,
)
def test_get_idps(urlopen):
    assert ciecplib_utils.get_idps("something") == {
        "Institution 1": "https://inst1.test",
        "Institution 2": "https://inst2.test",
    }
    assert urlopen.called_with("something")


@mock.patch(
    "{0}.urlopen".format(urlreqmodname),
    return_value=RAW_INSTITUTION_LIST,
)
def test_get_idp_urls(_):
    assert ciecplib_utils.get_idp_urls("Institution 1") == (
        "https://inst1.test",
        "https://inst1.test",
    )


@mock.patch(
    "{0}.urlopen".format(urlreqmodname),
    return_value=RAW_INSTITUTION_LIST,
)
@pytest.mark.parametrize("inst", ["Institution*", "something else"])
def test_get_idp_urls_error(_, inst):
    with pytest.raises(ValueError):
        ciecplib_utils.get_idp_urls(inst)


@pytest.mark.parametrize("url, result", [
    ("login.ligo.org",
     "https://login.ligo.org/idp/profile/SAML2/SOAP/ECP"),
    ("https://login.ligo.org",
     "https://login.ligo.org/idp/profile/SAML2/SOAP/ECP"),
    ("login.ligo.org/test",
     "https://login.ligo.org/test"),
    ("https://login.ligo.org/idp/profile/SAML2/SOAP/ECP",
     "https://login.ligo.org/idp/profile/SAML2/SOAP/ECP"),
])
def test_format_endpoint_url(url, result):
    assert ciecplib_utils.format_endpoint_url(url) == result
